"""Common shared utility classes and methods."""

from .template import render_template, render_recursive_template, render_template_inplace

__all__ = ['render_template', 'render_recursive_template', 'render_template_inplace']
